import os
import subprocess
import itertools
import tempfile
import time
import sys
import textwrap

import pytest

from yg.lockfile import FileLock, FileLockTimeout


def test_FileLock_basic():
    tfile, filename = tempfile.mkstemp()
    os.close(tfile)
    os.remove(filename)
    l = FileLock(filename)
    l2 = FileLock(filename, timeout=0.2)
    assert not l.is_locked()
    l.acquire()
    assert l.is_locked()
    l.release()
    assert not l.is_locked()
    with l:
        assert os.path.isfile(filename)
        with pytest.raises(FileLockTimeout):
            l2.acquire()
    assert not l.is_locked()
    l2.acquire()
    assert l2.is_locked()
    l2.release()

def lines(stream):
    """
    I can't figure out how to get the subprocess module to feed me
    line-buffered output from a sub-process, so I grab the output byte
    by byte and assemble it into lines.
    """
    buf = b''
    while True:
        dat = stream.read(1)
        if dat:
            buf += dat
            if dat == b'\n':
                yield buf
                buf = b''
        if not dat and buf:
            yield buf
        if not dat:
            break

def decoded_lines(stream):
    return (line.decode('utf-8-sig') for line in lines(stream))

def _eggs_env():
    """
    When running under setup.py test, dependencies are loaded dynamically.
    As a result, a subprocess will not have the necessary dependencies.
    Return an environment in which those dynamic dependencies will be
    included.
    """
    local_eggs = [egg for egg in sys.path if '.eggs' in egg]
    return dict(PYTHONPATH=os.pathsep.join(local_eggs))

def test_FileLock_process_killed():
    """
    If a subprocess fails to release the lock, it should be released
    and available for another process to take it.
    """
    tfile, filename = tempfile.mkstemp()
    os.close(tfile)
    os.remove(filename)
    script = textwrap.dedent("""
        from __future__ import print_function
        from yg.lockfile import FileLock
        import time
        l = FileLock({filename!r})
        l.acquire()
        print("acquired", l.lockfile)
        [time.sleep(1) for x in range(10)]
        """).format(**locals())
    script_lines = script.strip().split('\n')
    script_cmd = '; '.join(script_lines)
    cmd = [sys.executable, '-u', '-c', script_cmd]
    proc = subprocess.Popen(cmd, stdout=subprocess.PIPE, env=_eggs_env())
    lines = decoded_lines(proc.stdout)
    out = itertools.takewhile(lambda l: 'acquired' not in l, lines)
    tuple(out) # wait for 'acquired' to be printed by subprocess

    l = FileLock(filename, timeout=0.2)
    with pytest.raises(FileLockTimeout):
        l.acquire()
    proc.kill()
    time.sleep(.5)
    l.acquire()
    l.release()



#!/usr/bin/env python

"""
Cross-platform file lock context
"""

import os
import time
import functools
import numbers
import datetime
import logging

import zc.lockfile
from jaraco import timing

from . import py33compat


# disable the error reporting in zc.lockfile because it will
#  just spew errors as we repeatedly attempt to acquire an
#  unavailable lock.
logging.getLogger('zc.lockfile').setLevel(logging.ERROR+1)


class FileLockTimeout(Exception):
    pass

class LockBase(object):
    def __init__(self, timeout=10, delay=.05):
        if isinstance(timeout, numbers.Number):
            timeout = datetime.timedelta(seconds=timeout)
        if isinstance(delay, numbers.Number):
            delay = datetime.timedelta(seconds=delay)
        self.timeout = timeout
        self.delay = delay

    def __enter__(self):
        """
        Acquire the lock unless already locked.
        """
        if not self.is_locked():
            self.acquire()
        return self

    def __exit__(self, type, value, traceback):
        self.release()

    def __del__(self):
        """
        Release the lock on destruction.
        """
        self.release()


class FileLock(LockBase):
    """
    A cross-platform locking file context.

    May be used in a with statement to provide system-level concurrency
    protection.

    This class relies on zc.lockfile for the underlying locking.

    This class is not threadsafe.
    """

    def __init__(self, lockfile, *args, **kwargs):
        """
        Construct a FileLock. Specify the path to the file to lock and
        optionally the maximum timeout and the delay between each attempt to
        lock.

        Timeout and delay can be given in numeric seconds or as
        `datetime.timedelta` objects.
        """
        self.lockfile = lockfile
        super(FileLock, self).__init__(*args, **kwargs)

    def acquire(self):
        """
        Attempt to acquire the lock every `delay` seconds until the
        lock is acquired or until `timeout` has expired.

        Raises FileLockTimeout if the timeout is exceeded.

        Errors opening the lock file (other than if it exists) are
        passed through.
        """
        stopwatch = timing.Stopwatch()
        attempt = functools.partial(zc.lockfile.LockFile, self.lockfile)
        while True:
            try:
                self.lock = attempt()
                break
            except zc.lockfile.LockError:
                timeout_expired = stopwatch.split() >= self.timeout
                if timeout_expired:
                    raise FileLockTimeout()
                time.sleep(self.delay.total_seconds())

    def is_locked(self):
        return hasattr(self, 'lock')

    def release(self):
        """
        Release the lock and delete the lockfile.
        """
        if self.is_locked():
            self.lock.close()
            del self.lock
            with py33compat.suppress_file_not_found():
                os.remove(self.lockfile)


class ExclusiveContext(LockBase):
    """
    An exclusive context on an existing, open file.
    """
    def __init__(self, file, *args, **kwargs):
        """
        file should be an open file
        """
        self.file = file
        super(ExclusiveContext, self).__init__(*args, **kwargs)

    def acquire(self):
        """
        Attempt to acquire the lock every `delay` seconds until the
        lock is acquired or until `timeout` has expired.

        Raises FileLockTimeout if the timeout is exceeded.

        Errors opening the lock file (other than if it exists) are
        passed through.
        """
        stopwatch = timing.Stopwatch()
        attempt = functools.partial(zc.lockfile._lock_file, self.file)
        while True:
            try:
                self.lock = attempt()
                break
            except zc.lockfile.LockError:
                timeout_expired = stopwatch.split() >= self.timeout
                if timeout_expired:
                    raise FileLockTimeout()
                time.sleep(self.delay.total_seconds())

    def is_locked(self):
        return hasattr(self, 'lock')

    def release(self):
        """
        Release the lock and delete the lockfile.
        """
        if self.is_locked():
            zc.lockfile._unlock_file(self.file)
            del self.lock



import functools
import contextlib


try:
	suppress_file_not_found = functools.partial(
		contextlib.suppress,
		FileNotFoundError,
	)
except Exception:
	# Python 3.3 doesn't have contextlib.suppress
	# nor FileNotFoundError.
	import contextlib2
	suppress_file_not_found = functools.partial(
		contextlib2.suppress,
		OSError,
	)



#!/usr/bin/env python

# Project skeleton maintained at https://github.com/jaraco/skeleton

import io
import sys

import setuptools

with io.open('README.rst', encoding='utf-8') as readme:
	long_description = readme.read()

needs_wheel = {'release', 'bdist_wheel', 'dists'}.intersection(sys.argv)
wheel = ['wheel'] if needs_wheel else []

name = 'yg.lockfile'
description = 'Lockfile object with timeouts and context manager'

py33_markers = [
	'python_version=="{ver}"'.format(**locals())
	for ver in ('3.3', '3.2', '2.7')
]
py33 = ':' + ' or '.join(py33_markers)

setup_params = dict(
	name=name,
	use_scm_version=True,
	author="Jason R. Coombs",
	author_email="jaraco@jaraco.com",
	description=description or name,
	long_description=long_description,
	url="https://github.com/yougov/" + name,
	packages=setuptools.find_packages(),
	include_package_data=True,
	namespace_packages=name.split('.')[:-1],
	install_requires=[
		'zc.lockfile',
		'jaraco.timing',
	],
	extras_require={
		py33: [
			"contextlib2>=0.5",
		],
	},
	setup_requires=[
		'setuptools_scm>=1.15.0',
	] + wheel,
	classifiers=[
		"Development Status :: 5 - Production/Stable",
		"Intended Audience :: Developers",
		"License :: OSI Approved :: MIT License",
		"Programming Language :: Python :: 2.7",
		"Programming Language :: Python :: 3",
	],
	entry_points={
	},
)
if __name__ == '__main__':
	setuptools.setup(**setup_params)
